#include <assert.h>
#include "Constraint.h"
#include "mtl/Map.h"

using namespace std;
using namespace Cosoco;


// Constructors and delayed initialisation

Constraint::Constraint(Problem &p, std::string n, vec<Variable *> &vars) : indexesAreValues(true), problem(p), solver(nullptr),
                                                                           name(n),
                                                                           unassignedVariablesIdx(vars.size(), true),
                                                                           timestamp(0),
                                                                           wdeg(0) {
    scopeInitialisation(vars);

}


Constraint::Constraint(Problem &p, std::string n, Variable *xx, Variable *yy) : problem(p), solver(nullptr), name(n),
                                                                                unassignedVariablesIdx(2, true),
                                                                                timestamp(0), wdeg(0) {
    vec<Variable *> v;
    v.push(xx);
    v.push(yy);
    scopeInitialisation(v);
}


Constraint::Constraint(Problem &p, std::string n, Variable *x) : problem(p), solver(nullptr), name(n),
                                                                 unassignedVariablesIdx(1, true),
                                                                 timestamp(0), wdeg(0) {
    vec<Variable *> v;
    v.push(x);
    scopeInitialisation(v);
}


Constraint::Constraint(Problem &p, std::string n, Variable *x, Variable *y, Variable *z) : problem(p), solver(nullptr), name(n),
                                                                                           unassignedVariablesIdx(3,
                                                                                                                  true),
                                                                                           timestamp(0),
                                                                                           wdeg(0) {
    vec<Variable *> v;
    v.push(x);
    v.push(y);
    v.push(z);
    scopeInitialisation(v);
}


Constraint::Constraint(Problem &p, std::string n, int sz) : problem(p), solver(nullptr), name(n), unassignedVariablesIdx(sz, true),
                                                            timestamp(0), wdeg(0) {}


void Constraint::scopeInitialisation(vec<Variable *> &vars) {

    assert(scope.size() == 0); // Do it only one
    vars.copyTo(scope);
    idxToScopePosition.growTo(problem.variables.size(), NOTINSCOPE);
    arity = scope.size();
    indexesAreValues = true;
    for(int i = 0; i < vars.size(); i++) {
        indexesAreValues = indexesAreValues &&
                           (vars[i]->domain[0] == 0 && vars[i]->domain[vars[i]->size() - 1] == vars[i]->size() - 1);
        idxToScopePosition[vars[i]->idx] = i;
    }
    // TODO a bug occurs here
    indexesAreValues = false;
    current.growTo(vars.size());
}


bool Constraint::isCorrectlyDefined() {
    return true;
}


bool Constraint::scopeIsOk() {
    for(Variable *x : scope)
        x->fake = 0;
    for(Variable *x : scope) {
        if(x->fake == 1)
            throw std::logic_error("Constraint " + std::to_string(idc) + ": scope contains variable " + x->name() + " many times");
        x->fake = 1;
    }
    return true;
}


void Constraint::delayedConstruction(int id) {
    assert(scope.size() > 0); // scopeInitialisation have to be done
    idc = id;
    for(Variable *v : scope)
        v->addConstraint(this);
}


void Constraint::attachSolver(Solver *s) {
    solver = s;
}


// Clauses that can be derived initialy by the constraint in order to help the csp solver

void Constraint::addInitialSATClauses() {} // Nothing to do


// Filtering

State Constraint::status() { return UNDEF; }   // By default, a constraint must be filtering at each call

void Constraint::reinitialize() {}          // By default, nothing to do

bool Constraint::filterFrom(Variable *x) {
    State st = status();
    if(st == CONSISTENT) return true;
    if(st == INCONSISTENT) return false;
    return filter(x);
}


// Assign and unassign variables
void Constraint::assignVariable(Variable *x) {
    int posx = toScopePosition(x);
    assert(unassignedVariablesIdx.contains(posx));
    unassignedVariablesIdx.del(posx);
}


void Constraint::unassignVariable(Variable *x) {
    int posInScope = toScopePosition(x);
    assert(!unassignedVariablesIdx.contains(posInScope));
    unassignedVariablesIdx.add(posInScope);
}


// idx -> scope position
int Constraint::toScopePosition(int idx) {
    switch(scope.size()) {
        case 2 :
            return scope[0]->idx == idx ? 0 : 1;
        case 3 :
            if(scope[0]->idx == idx) return 0;
            if(scope[1]->idx == idx) return 1;
            return 2;
            break;
    }
    return idxToScopePosition[idx];
}


int Constraint::toScopePosition(Variable *x) {
    return toScopePosition(x->idx);
}


bool Constraint::isSatisfiedByOfIndexes(vec<int> &tupleOfIndex) {
    if(indexesAreValues)
        return isSatisfiedBy(tupleOfIndex);
    for(int i = 0; i < scope.size(); i++)
        current[i] = scope[i]->domain.toVal(tupleOfIndex[i]);

    return isSatisfiedBy(current);
}


void Constraint::extractConstraintTupleFromInterpretation(const vec<int> &interpretation, vec<int> &tuple) {
    tuple.clear();
    for(int i = 0; i < scope.size(); i++) {
        int idx = scope[i]->idx;
        tuple.push(interpretation[idx]);
    }


}


Constraint *Constraint::toExtensionConstraint() {
    //
    int nbTuples = 1;
    for(Variable *x : scope)
        nbTuples *= x->domain.maxSize();
    if(nbTuples > 100000) // To many tuples
        return nullptr;

    vec<vec<int> > conflicts, supports;
    vec<int> tuple;
    tuple.growTo(scope.size());
    createTuples(0, conflicts, supports, tuple);

    display();
    std::cout << conflicts.size() << "  " << supports.size() << std::endl;
/*    if(conflicts.size() < supports.size())
        return FactoryConstraints::newExtensionConstraint(&problem, "", scope, conflicts, false);
    else
        return FactoryConstraints::newExtensionConstraint(&problem, "", scope, supports, true);
*/
    return nullptr;
 }


void Constraint::createTuples(int posx, vec<vec<int>> &conflicts, vec<vec<int>> &supports, vec<int> &tuple) {
    Variable *x = scope[posx];
    for(int i = 0; i < x->domain.maxSize(); i++) {
        tuple[posx] = x->domain.toVal(i);
        if(posx == scope.size() - 1) {
            vec<vec<int>> &putInside = isSatisfiedBy(tuple) ? supports : conflicts;
            putInside.push();
            tuple.copyTo(putInside.last());
        } else
            createTuples(posx + 1, conflicts, supports, tuple);
    }
}


// Display
void Constraint::display(bool allDetails) {
    printf("Constraint {%s} %s : [", (typeid(*this).name()) + 9, name.c_str()); // 9 to remove namespace...
    for(int i = 0; i < scope.size(); i++) printf(" %s ", scope[i]->name());
    printf("]\n");
}      

