#ifndef FACTORYCONSTRAINTS_H
#define    FACTORYCONSTRAINTS_H

#include "Constraint.h"
#include "mtl/Vec.h"
#include "extensions/BinaryExtension.h"
#include "extensions/ShortSTR2.h"
#include "extensions/Unary.h"
#include "extensions/STRNeg.h"
#include "extensions/MDDExtension.h"
#include "globals/comparison/AllDifferent.h"
#include "globals/comparison/DistinctVectors.h"
#include "globals/comparison/NotAllEqual.h"
#include "primitives/EQ.h"
#include "primitives/Disjunctive.h"
#include "globals/summing/Sum.h"
#include "globals/summing/SumEQ.h"
#include "globals/summing/SumGE.h"
#include "globals/summing/SumNE.h"
#include "globals/connection/element/ElementConstant.h"
#include "globals/connection/element/ElementVariable.h"
#include "globals/connection/maximum/MaximumConstantLE.h"
#include "globals/connection/maximum/MaximumConstantGE.h"
#include "globals/connection/minimum/MinimumConstantLE.h"
#include "globals/connection/minimum/MinimumConstantGE.h"
#include "globals/connection/minimum/MinimumVariableEQ.h"
#include "constraints/globals/connection/maximum/MaximumVariableEQ.h"
#include "globals/counting/AtLeastK.h"
#include "globals/counting/AtMostK.h"
#include "globals/counting/ExactlyK.h"
#include "globals/counting/NValuesLEK.h"
#include "globals/counting/ExactlyKVariable.h"
#include "intension/Intension.h"
#include "globals/comparison/Lexicographic.h"
#include "genericFiltering/AC3rm.h"
#include "primitives/LE.h"
#include "primitives/LT.h"
#include "primitives/DiffXY.h"
#include "XCSP3Constants.h"
#include "utils/Verbose.h"
#include <regex>
#include <iostream>

namespace Cosoco {
    using namespace XCSP3Core;

    typedef struct Occurs Occurs;
    struct Occurs {
        int value;   //One of value (min,max) or Variables are uses
        int min, max;
        Variable *x;
        int type;
    };

    enum {
        OCCURS_INTEGER, OCCURS_INTERVAL, OCCURS_VARIABLE
    };

    static Verbose verbose;

    class FactoryConstraints {
    public :
        //--------------------------------------------------------------------------------------
        // Basic constraints
        //--------------------------------------------------------------------------------------

        static void createConstraintIntension(Problem *p, std::string name, XCSP3Core::Tree *tree, vec<Variable *> &scope) {
            p->addConstraint(new AdapterAC3rm(new Intension(*p, name, tree, scope)));
        }

        static void createConstraintLessThan(Problem *p, std::string name, Variable *x, int k, Variable *y, bool strict) {
            if(strict)
                p->addConstraint(new Lt(*p, name, x, y, k));
                else
            p->addConstraint(new Le(*p, name, x, y, k));
        }

        static Constraint *
        newExtensionConstraint(Problem *p, std::string name, vec<Variable *> &vars, vec<vec<int> > &tuples, bool isSupport, bool hasStar = false) {
            Extension *ctr = nullptr;
            if(vars.size() == 2)
                ctr = new BinaryExtension(*p, name, isSupport, vars[0], vars[1]);
            else {
                if(isSupport)
                    ctr = new ShortSTR2(*p, name, vars);
                else {
                    assert(hasStar == false); // TODO
                    ctr = new STRNeg(*p, name, vars);
                }
            }

            for(int i = 0 ; i < tuples.size() ; i++)
                ctr->addTuple(tuples[i]);
            return ctr;
        }


        static void
        createConstraintExtension(Problem *p, std::string name, vec<Variable *> &vars, vec<vec<int> > &tuples, bool isSupport, bool hasStar = false) {
            p->addConstraint(newExtensionConstraint(p, name, vars, tuples, isSupport, hasStar));
        }


        static void createConstraintExtensionAs(Problem *p, std::string name, vec<Variable *> &vars, bool isSupport) {
            Extension *ctr = nullptr;
            Extension *sameConstraint = (Extension *) p->constraints.last();
            assert(sameConstraint->scope.size() == vars.size());

            Extension::nbShared++;

            if(vars.size() == 1) {
                p->addConstraint(new Unary(*p, name, vars[0], ((Unary *) p->constraints.last())->values, isSupport));
                return;
            }

            if(vars.size() == 2)
                ctr = new BinaryExtension(*p, name, isSupport, vars[0], vars[1], (BinaryExtension *) sameConstraint);

            if(vars.size() > 2) {
                if(isSupport)
                    ctr = new ShortSTR2(*p, name, vars, sameConstraint->tuples);
                else
                    ctr = new STRNeg(*p, name, vars, sameConstraint->tuples);
            }
            p->addConstraint(ctr);
        }


        static void createConstraintUnary(Problem *p, std::string name, Variable *x, vec<int> &values, bool isSupport) {
            p->addConstraint(new Unary(*p, name, x, values, isSupport));
        }

        //--------------------------------------------------------------------------------------
        // Language constraints
        //--------------------------------------------------------------------------------------

        static void createConstraintMDD(Problem *p, std::string name, vec<Variable *> &vars, vec<XTransition *> &transitions) {
            p->addConstraint(new MDDExtension(*p, name, vars, transitions));
        }


        static void createConstraintMDD(Problem *p, std::string name, vec<Variable *> &vars, MDD *mdd) {
            p->addConstraint(new MDDExtension(*p, name, vars, mdd));
        }


        static void createConstraintRegular(Problem *p, std::string name, vec<Variable *> &vars, string start, std::vector<string> &final,
                                            vec<XTransition *> &transitions) {
            p->addConstraint(new MDDExtension(*p, name, vars, MDD::buildFromAutomata(name, vars, start, final, transitions)));
        }

        //--------------------------------------------------------------------------------------
        // Comparison constraints
        //--------------------------------------------------------------------------------------

        static void createConstraintAllDiff(Problem *p, std::string name, vec<Variable *> &vars) {
            if(vars.size() == 2)
                p->addConstraint(new DiffXY(*p, name, vars[0], vars[1]));
            else
                p->addConstraint(new AllDifferent(*p, name, vars));
        }


        static void createConstraintAllDiffExcept(Problem *p, std::string name, vec<Variable *> &vars, vector<int> &except) {
            for(int i = 0 ; i < vars.size() ; i++) {
                Variable *x1 = vars[i];
                for(int j = 0 ; j < i ; j++) {
                    Variable *x2 = vars[j];
                    vec<vec<int> > tuples;
                    std::cout << x2->size() << " " << x1->size() << std::endl;
                    for(int idv1 : x1->domain) {
                        int v1 = x1->domain.toVal(idv1);
                        for(int idv2 : x2->domain) {
                            int v2 = x2->domain.toVal(idv2);
                            if(v1 != v2) {
                                tuples.push();
                                vec<int> &t = tuples.last();
                                t.push(idv1);
                                t.push(idv2);
                            }
                        }
                    }
                    for(int v : except) {
                        std::cout << x1->name() << " " << x2->name() << std::endl;
                        if(x1->containsValue(v)) {
                            tuples.push();
                            vec<int> &t = tuples.last();
                            t.push(x1->domain.toIdv(v));
                            t.push(STAR);
                        }
                        if(x2->containsValue(v)) {
                            tuples.push();
                            vec<int> &t = tuples.last();
                            t.push(STAR);
                            t.push(x2->domain.toIdv(v));
                        }
                    }
                    vec<Variable *> varsInConstraint;
                    varsInConstraint.push(x1);
                    varsInConstraint.push(x2);
                    createConstraintExtension(p, name, varsInConstraint, tuples, true, true);
                }
            }
        }

        //----------------------------------------------------------------------


        static void createExtenstionDistinctVector(Problem *p, std::string name, vec<Variable *> &list1, vec<Variable *> &list2) {

            vec<Variable *> scope;
            vec<int> position;
            position.growTo(p->nbVariables(), -1);

            for(Variable *x : list1) {
                if(position[x->idx] == -1) {
                    scope.push(x);
                    position[x->idx] = scope.size() - 1;
                }
            }
            for(Variable *x : list2) {
                if(position[x->idx] == -1) {
                    scope.push(x);
                    position[x->idx] = scope.size() - 1;
                }
            }

            vec<vec<int> > tuples;

            for(int i = 0 ; i < list1.size() ; i++) {
                Variable *x = list1[i];
                Variable *y = list2[i];
                if(x == y)
                    continue;
                for(int v1 : x->domain) {
                    for(int v2 : y->domain) {
                        if(v1 == v2)
                            continue;
                        tuples.push();
                        vec<int> &tp = tuples.last();
                        tp.growTo(scope.size(), STAR);
                        tp[position[x->idx]] = x->domain.toIdv(v1);
                        tp[position[y->idx]] = y->domain.toIdv(v2);
                    }
                }
            }
            createConstraintExtension(p, name, scope, tuples, true, true);
        }


        static void createConstraintAllDiffList(Problem *p, std::string name, vec<vec<Variable *> > &lists) {
            if(0) {
                verbose.log(NORMAL, "c AllDiff List constraint using %d DistinctVectors constraints \n",
                            lists.size() * (lists.size() - 1) / 2);

                for(int i = 0 ; i < lists.size() ; i++)
                    for(int j = i + 1 ; j < lists.size() ; j++)
                        p->addConstraint(new DistinctVectors(*p, name, lists[i], lists[j]));
            } else {
                verbose.log(NORMAL, "c AllDiff List constraint using extension constraint\n", lists.size() * (lists.size() - 1) / 2);
                for(int i = 0 ; i < lists.size() ; i++)
                    for(int j = i + 1 ; j < lists.size() ; j++)
                        createExtenstionDistinctVector(p, name, lists[i], lists[j]);
            }
        }



        //----------------------------------------------------------------------


        static void createConstraintAllEqual(Problem *p, std::string name, vec<Variable *> &vars) {
            verbose.log(NORMAL, "c AllEqual constraint using %d generic x=y constraints \n", vars.size() - 1);
            for(int i = 0 ; i < vars.size() - 1 ; i++)
                p->addConstraint(new EQ(*p, name, vars[i], vars[i + 1]));
        }


        static void createConstraintNotAllEqual(Problem *p, std::string name, vec<Variable *> &vars) {
            p->addConstraint(new NotAllEqual(*p, name, vars));
        }


        //--------------------------------------------------------------------------------------
        // Summing and counting constraints
        //--------------------------------------------------------------------------------------

        static void createConstraintSum(Problem *p, std::string name, vec<Variable *> &vars, vec<int> &coeffs, long l, OrderType order) {
            Sum *ctr = nullptr;

            // Rearrange coeffs.
            if(order == OrderType::LE || order == OrderType::LT) {
                for(int i = 0 ; i < coeffs.size() ; i++)
                    coeffs[i] = -coeffs[i];
                l = -l;
            }

            // Need order on coefficients
            for(int i = 0 ; i < vars.size() ; i++) {
                int pos = i;
                for(int j = i + 1 ; j < vars.size() ; j++) {
                    if(coeffs[j] < coeffs[pos])
                        pos = j;
                }
                int tmp = coeffs[i];
                coeffs[i] = coeffs[pos];
                coeffs[pos] = tmp;
                Variable *x = vars[i];
                vars[i] = vars[pos];
                vars[pos] = x;
            }

            switch(order) {
                case OrderType::LE :
                    ctr = new SumGE(*p, name, vars, coeffs, l);
                    break;
                case OrderType::LT :
                    ctr = new SumGE(*p, name, vars, coeffs, l + 1);
                    break;
                case OrderType::GE :
                    ctr = new SumGE(*p, name, vars, coeffs, l);
                    break;
                case OrderType::GT :
                    ctr = new SumGE(*p, name, vars, coeffs, l + 1); //TODO
                    break;
                case OrderType::IN :
                    throw runtime_error("This is forbidden to construct a sum with IN operator");
                    break;
                case OrderType::EQ:
                    ctr = new SumEQ(*p, name, vars, coeffs, l);
                    break;
                case OrderType::NE :
                    ctr = new SumNE(*p, name, vars, coeffs, l);
                    break;

            }
            assert(ctr != nullptr);
            p->addConstraint(ctr);
        }


        static void createConstraintAtLeast(Problem *p, std::string name, vec<Variable *> &vars, int value, int k) {
            p->addConstraint(new AtLeastK(*p, name, vars, k, value));
        }


        static void createConstraintAtMost(Problem *p, std::string name, vec<Variable *> &vars, int value, int k) {
            p->addConstraint(new AtMostK(*p, name, vars, k, value));
        }


        static void createConstraintExactly(Problem *p, std::string name, vec<Variable *> &vars, int value, int k) {
            p->addConstraint(new ExactlyK(*p, name, vars, k, value));
        }


        static void createConstraintExactlyVariable(Problem *p, std::string name, vec<Variable *> &vars, int value, Variable *k) {
            p->addConstraint(new ExactlyKVariable(*p, name, vars, k, value));
        }

        static void createConstraintNValuesLE(Problem *p, std::string name, vec<Variable *> &vars, int k) {
            p->addConstraint(new NValuesLEK(*p, name, vars, k));
        }


        //--------------------------------------------------------------------------------------
        // Connection constraints
        //--------------------------------------------------------------------------------------
        static void createConstraintElementConstant(Problem *p, std::string name, vec<Variable *> &vars, Variable *index, int startIndex, int v) {
            p->addConstraint(new ElementConstant(*p, name, vars, index, v, startIndex == 1));
        }


        static void createConstraintElementVariable(Problem *p, std::string name, vec<Variable *> &vars, Variable *index, int startIndex, Variable *v) {
            p->addConstraint(new ElementVariable(*p, name, vars, index, v, startIndex == 1));
        }


        static void createConstraintCardinality(Problem *p, std::string name, vec<Variable *> &vars, vec<int> &values, vec<Occurs> &occurs) {
            verbose.log(NORMAL, "c Create Cardinality(vars,values,occurs) constraint using %d Exactly/AtMost.. constraint \nc\n", vars.size());
            if(occurs.size() != values.size())
                throw std::logic_error("Cardinality: Occurs and values must have the same size");
            for(int i = 0 ; i < occurs.size() ; i++) {
                if(occurs[i].type == OCCURS_INTEGER)
                    createConstraintExactly(p, name + "card as exactly", vars, values[i], occurs[i].value);
                if(occurs[i].type == OCCURS_INTERVAL) {
                    createConstraintAtLeast(p, name + "card as atleast", vars, values[i], occurs[i].min);
                    createConstraintAtMost(p, name + "card as atleast", vars, values[i], occurs[i].max);
                }
                if(occurs[i].type == OCCURS_VARIABLE) {
                    createConstraintExactlyVariable(p, name + "card as exactly", vars, values[i], occurs[i].x);
                }
            }
        }


        static void createConstraintOrdered(Problem *p, std::string name, vec<Variable *> &vars, vector<int> &lengths, OrderType op) {
            for(int i = 0 ; i < vars.size() - 1 ; i++) {
                int k = lengths.size() == 0 ? 0 : lengths[i];
                std::cout << k << std::endl;
                if(op == OrderType::LE)
                    p->addConstraint(new Le(*p, name + vars[i]->name() + " op " + name + vars[i + 1]->name(), vars[i], vars[i + 1], k));
                if(op == OrderType::LT)
                    p->addConstraint(new Lt(*p, name + vars[i]->name() + " op " + name + vars[i + 1]->name(), vars[i], vars[i + 1], k));

                if(op == OrderType::GE)
                    p->addConstraint(
                            new Le(*p, name + vars[i]->name() + " op " + name + vars[i + 1]->name(), vars[i + 1], vars[i], -k));
                if(op == OrderType::GT)
                    p->addConstraint(
                            new Lt(*p, name + vars[i]->name() + " op " + name + vars[i + 1]->name(), vars[i + 1], vars[i], -k));
            }
        }


        static void createConstraintLex(Problem *p, std::string name, vec<Variable *> &vars1, vec<Variable *> &vars2, OrderType op) {
            if(op == OrderType::LE)
                p->addConstraint(new Lexicographic(*p, name, vars1, vars2, false));
            if(op == OrderType::LT)
                p->addConstraint(new Lexicographic(*p, name, vars1, vars2, true));

            if(op == OrderType::GE)
                p->addConstraint(new Lexicographic(*p, name, vars2, vars1, false));
            if(op == OrderType::GT)
                p->addConstraint(new Lexicographic(*p, name, vars2, vars1, true));
        }


        static void createContraintChannelXY(Problem *p, std::string name, vec<Variable *> &X, vec<Variable *> &Y, int startX, int startY) {
            verbose.log(NORMAL, "c Create channel(X,Y) constraint using %d element constraint \n", X.size());
            for(int i = 0 ; i < X.size() ; i++)
                createConstraintElementConstant(p, name, Y, X[i], startY, i);
        }

        //-----------------------------------------------------------------------

        static void createConstraintNoOverlap(Problem *p, std::string name, Variable *x1, Variable *x2, int w1, int w2) {
            p->addConstraint(new Disjunctive(*p, name, x1, x2, w1, w2));
            //string tmp = "or(le(add(" + x1->_name + "," + to_string(w1) + ")," + x2->_name + "),le(add(" + x2->_name + "," + to_string(w2) + ")," + x1->_name + "))";
            //createConstraintIntension(p, "no overlap", tmp);
        }

        //-----------------------------------------------------------------------

        static void createConstraintMaximumLE(Problem *p, std::string name, vec<Variable *> &vars, int k) {
            p->addConstraint(new MaximumConstantLE(*p, name, vars, k));
        }


        static void createConstraintMaximumGE(Problem *p, std::string name, vec<Variable *> &vars, int k) {
            p->addConstraint(new MaximumConstantGE(*p, name, vars, k));
        }


        static void createConstraintMinimumLE(Problem *p, std::string name, vec<Variable *> &vars, int k) {
            p->addConstraint(new MinimumConstantLE(*p, name, vars, k));
        }


        static void createConstraintMinimumGE(Problem *p, std::string name, vec<Variable *> &vars, int k) {
            p->addConstraint(new MinimumConstantGE(*p, name, vars, k));
        }


        static void createConstraintMinimumVariableEQ(Problem *p, std::string name, vec<Variable *> &vars, Variable *value) {
            if(vars.contains(value)) {
                verbose.log(NORMAL, "c min(x...)=x -> use inequalities\n");
                for(Variable *x : vars) {
                    if(x != value)
                        p->addConstraint(new Le(*p, name, value, x, 0));
                }
            } else
                p->addConstraint(new MinimumVariableEQ(*p, name, vars, value));
        }


        static void createConstraintMaximumVariableEQ(Problem *p, std::string name, vec<Variable *> &vars, Variable *value) {
            if(vars.contains(value)) {
                verbose.log(NORMAL, "c max(x...)=x -> use inequalities\n");
                for(Variable *x : vars) {
                    if(x != value)
                        p->addConstraint(new Le(*p, name, x, value, 0));
                }
            } else
                p->addConstraint(new MaximumVariableEQ(*p, name, vars, value));
        }

    };

}

#endif    /* FACTORYCONSTRAINTS_H */

