#include "constraints/Constraint.h"
#include "solver/Solver.h"
#include "Sum.h"
#include "SumGE.h"

using namespace Cosoco;



//----------------------------------------------
// Check validity and correct definition
//----------------------------------------------

bool SumGE::isSatisfiedBy(vec<int> &tuple) {
    return weightedSum(tuple) >= limit;
}

//----------------------------------------------
// Filtering
//----------------------------------------------

bool SumGE::filter(Variable *dummy) {
    computeBounds();
    if (min >= limit)
        return true;
    if (max < limit)
        return false;

    for (int i = right ; i >= leftmostPositiveCoefficientPosition ; i--) {
        Variable *x = scope[i];


        int sizeBefore = x->size();
        if (sizeBefore == 1)
            continue;

        long maxBase = max - x->maximum() * coefficients[i];
        int minimumBefore = x->minimum();

        for (int j = x->size() - 1 ; j >= 0 ; j--) {
            int v = x->domain.valueAtPosition(j);
            if (maxBase + v * coefficients[i] < limit && solver->delVal(x, v) == false)
                return false;
        }
        if ((min += (x->minimum() - minimumBefore) * coefficients[i]) >= limit)
            return true;
    }
    for (int i = left ; i < leftmostPositiveCoefficientPosition ; i++) {
        Variable *x = scope[i];

        int sizeBefore = x->size();
        if (sizeBefore == 1)
            continue;
        long maxBase = max - x->minimum() * coefficients[i];
        int maxBefore = x->maximum();

        for (int j = x->size() - 1 ; j >= 0 ; j--) {
            int v = x->domain.valueAtPosition(j);
            if (maxBase + v * coefficients[i] < limit && solver->delVal(x, v) == false)
                return false;
        }
        if (sizeBefore - x->size() > 0) {
            if ((min += (x->maximum() - maxBefore) * coefficients[i]) >= limit)
                return true;
        }
    }
    return true;
}


void SumGE::computeBounds() {
    min = max = 0;
    right = -1; // if it remains equal to -1, it indicates that no filtering has to be performed wrt positive coefficients
    for (int i = coefficients.size() - 1 ; i >= leftmostPositiveCoefficientPosition ; i--) { // positive coefs
        Domain &domain = scope[i]->domain;
        min += domain.minimum() * coefficients[i];
        max += domain.maximum() * coefficients[i];
        if (domain.size() > 1 && right == -1)
            right = i;
    }
    left = coefficients.size(); // if it remains equal to coefficients.length, it indicates that no filtering has to
    for (int i = 0 ; i < leftmostPositiveCoefficientPosition ; i++) { // Negative coefs
        Domain &domain = scope[i]->domain;
        min += domain.maximum() * coefficients[i];
        max += domain.minimum() * coefficients[i];
        if (domain.size() > 1 && left == coefficients.size())
            left = i;
    }
}



//----------------------------------------------
// Objective constraint
//----------------------------------------------

void SumGE::updateBound(long bound) {
    limit = bound;
}        // Update the current bound

long SumGE::maxUpperBound() {
    computeBounds();
    return max;
}

long SumGE::minLowerBound() {
    computeBounds();
    return min;
}


long SumGE::computeScore(vec<int> &solution) {
    return weightedSum(solution);
}

//----------------------------------------------
// Construction and initialisation
//----------------------------------------------


SumGE::SumGE(Problem &p, std::string n, vec<Variable *> &vars, vec<int> &coefs, long l) : Sum(p, n, vars, coefs, l) {
    leftmostPositiveCoefficientPosition = coefficients.size();
    for (int i = coefficients.size() - 1 ; i >= 0 ; i--) {
        if (coefficients[i] < 0)
            break;
        leftmostPositiveCoefficientPosition = i;
    }
}

