#include "constraints/primitives/xTimesyEQz.h"
#include "solver/Solver.h"

using namespace Cosoco;


//----------------------------------------------
// Check validity and correct definition
//----------------------------------------------

bool xTimesyEQz::isSatisfiedBy(vec<int> &tuple) {
    return tuple[0] * tuple[1] == tuple[2];
}


//----------------------------------------------
// Filtering
//----------------------------------------------

bool xTimesyEQz::filter(Variable *dummy) {

    // Filtering already done in special cases where x=0 or y=0
    if(solver->isAssigned(x) && x->value() == 0)
        return solver->assignToVal(z, 0);
    if(solver->isAssigned(y) && y->value() == 0)
        return solver->assignToVal(z, 0);

    int xmin = x->minimum(), xmax = x->maximum();
    int ymin = y->minimum(), ymax = y->maximum();

    // Update z bounds
    if(xmin >= 0 && ymin >= 0) { // z>=0
        solver->delValuesLowerOrEqualThan(z, xmin * ymin - 1);
        solver->delValuesGreaterOrEqualThan(z, xmax * ymax + 1);
    } else if(xmax < 0 && ymax < 0) { // Z>=0
        solver->delValuesLowerOrEqualThan(z, xmax * ymax - 1);
        solver->delValuesGreaterOrEqualThan(z, xmin * ymin + 1);
    } else if((xmin > 0 && ymax < 0) || (xmax < 0 && ymin > 0)) { // Z<0
        int a = xmin * ymax;
        int b = xmax * ymin;
        solver->delValuesLowerOrEqualThan(z, std::min(a, b) - 1);
        solver->delValuesGreaterOrEqualThan(z, std::max(a, b) + 1);
    }
    if(z->size() == 0) // Here to avoid to if()....)
        return false;

    if(solver->isAssigned(x) && instantiated(x, y) == false)
        return false;
    if(solver->isAssigned(y) && instantiated(y, x) == false)
        return false;
    assert(x->size() > 0);
    assert(y->size() > 0);
    assert(z->size() > 0);
    return true;

}


bool xTimesyEQz::instantiated(Variable *w, Variable *t) {
    int vw = w->value();
    if(solver->isAssigned(t))
        return solver->assignToVal(z, vw * t->value());
    if(solver->isAssigned(z)) {
        int approx = abs(z->value() / vw);
        return solver->assignToVal(y, approx);
    }
    int v1 = vw * t->minimum();
    int v2 = vw * t->maximum();
    if(solver->delValuesLowerOrEqualThan(z, std::min(v1, v2) - 1) == false)
        return false;
    return solver->delValuesGreaterOrEqualThan(z, std::max(v1, v2) + 1);
}

//----------------------------------------------
// Construction and initialisation
//----------------------------------------------

xTimesyEQz::xTimesyEQz(Problem &p, std::string n, Variable *xx, Variable *yy, Variable *zz) : Ternary(p, n, xx, yy, zz) {
    type = "(X * Y = Z)";
}



