#ifndef DOMAIN_H
#define DOMAIN_H

#include <assert.h>
#include "mtl/Vec.h"
#include "mtl/SparseSetMultiLevel.h"
#include "utils/Verbose.h"

namespace Cosoco {

    class Domain {
        friend class Variable;

        friend class Solver;

    protected :
        SparseSetMultiLevel sparseIdvs;  // the set of elements 
        int minval, maxval;
        bool checkMin, checkMax;
    public :
        using const_iterator = VecIterator<const int>;
        using iterator = VecIterator<int>;
        using const_reverse_iterator = VecIterator<const int>;
        using reverse_iterator = VecIterator<int>;

        // Constructors and initialisation
        Domain(int sz) : sparseIdvs(sz, true), checkMin(false), checkMax(false) { }


        Domain(Domain &d) : sparseIdvs(d.sparseIdvs) { }


        void delayedConstruction(int nbVars) { }


        // Virtual Method conversion id to value
        virtual int toIdv(int v) = 0;

        virtual int toVal(int idv) = 0;


        // Methods related to deletion of values
        inline void delIdv(int idv) { sparseIdvs.del(idv); }


        inline void delVal(int v) { sparseIdvs.del(toIdv(v)); } // TODO Check range !!!!!!


        inline void reduceTo(int idv) {
            sparseIdvs.clear();
            sparseIdvs.add(idv);
            minval = toVal(idv);
            maxval = toVal(idv);
        }

        bool isIdentical(Domain &d) {
            if(maxSize() != d.maxSize()) return false;
            for(int i = 0; i < maxSize(); i++)
                if(d.containsValue(toVal(sparseIdvs[i])) == false)
                    return false;
            return true;

        }

        void reinitialize() {  sparseIdvs.fill();  }


        inline int valueAtPosition(int pos) { return toVal(sparseIdvs[pos]); }


        inline const int getPosition(int idx) { return sparseIdvs.getPosition(idx); }


        // Method related to presence of value
        inline bool containsValue(int v) {
            int idv = toIdv(v);
            return idv == -1 ? false : sparseIdvs.contains(idv);
        }


        inline bool containsIdv(int idv) { return sparseIdvs.contains(idv); }


        inline bool isUniqueValue(int v) { return size() == 1 && containsValue(v); }


        inline int minimum() {
            assert(size() > 0);
            if(checkMin || containsValue(minval) == false) {
                minval = toVal(sparseIdvs[0]);
                for(int idv : sparseIdvs)
                    if(toVal(idv) < minval)
                        minval = toVal(idv);
                checkMin = false;
            }
            return minval;
        }


        inline int maximum() {
            assert(size() > 0);
            if(checkMax || containsValue(maxval) == false) {
                maxval = toVal(sparseIdvs[0]);
                for(int idv : sparseIdvs)
                    if(toVal(idv) > maxval)
                        maxval = toVal(idv);
                checkMax = false;
            }
            return maxval;
        }


        // return the next valid index in the domain, -1 if none exists (in such a case, idx is equal to the max
        inline int nextValidIndex(int currentIdv) {
            assert(size() > 0);
            int m = maximum();
            if(toIdv(m) == currentIdv) return -1;
            for(int idv = currentIdv + 1; idv < maxSize(); idv++) {
                if(containsIdv(idv))
                    return idv;
            }
            return -1;
        }


        // Methods related to size
        inline bool isEmpty() { return sparseIdvs.isEmpty(); }


        inline int size() { return sparseIdvs.size(); }


        inline const int maxSize() { return sparseIdvs.maxSize(); }


        // Access method
        inline int operator[](const int i) { return sparseIdvs[i]; }


        // begin/end functions to use for each aka c++
        iterator begin() { return sparseIdvs.begin(); }
        const_iterator cbegin() const { return sparseIdvs.cbegin(); }
        iterator end() { return sparseIdvs.end(); }
        const_iterator cend() const { return sparseIdvs.cend(); }

        // rbegin/rend functions to reverse traversal
        reverse_iterator rbegin() { return sparseIdvs.rbegin(); }
        const_reverse_iterator rbegin() const { return sparseIdvs.crbegin(); }
        reverse_iterator rend() { return sparseIdvs.rend(); }
        const_reverse_iterator rend() const { return sparseIdvs.crend(); }




        // Methods related to Multilevel of sparseset
        inline void recordLimit(int level) {  sparseIdvs.recordLimit(level); }


        inline void restoreLimit(int level) {
            sparseIdvs.restoreLimit(level);
            checkMin = true;
            checkMax = true;

        }


        // Display
        virtual void display() {
            printf("{");
            for(int idv = 0; idv < sparseIdvs.maxSize(); idv++) {
                if(idv >= size()) printf("%s", KRED);
                printf(" %d ", toVal(sparseIdvs[idv]));
                printf("%s", KNRM);
            }
            printf("}");

        }


    };

};

#endif
