#ifndef COSOCO_COSOCOCALLBACKS_H
#define COSOCO_COSOCOCALLBACKS_H

#include "XCSP3CoreCallbacks.h"


#include "XCSP3Constants.h"
#include "core/Problem.h"
#include "core/OptimizationProblem.h"
#include "core/domain/DomainRange.h"
#include "core/domain/DomainValues.h"
#include "constraints/FactoryConstraints.h"
#include "XCSP3Constraint.h"
#include "constraints/extensions/structures/MDD.h"

using namespace Cosoco;
using namespace std;

namespace XCSP3Core {
    class CosocoCallbacks : public XCSP3CoreCallbacks {
    protected :

        vec<int> &vector2vec(vector<int> &src) {
            vals.clear();
            vals.growTo(src.size());
            for(unsigned int i = 0 ; i < src.size() ; i++)
                vals[i] = src[i];
            return vals;
        }


        vec<Variable *> &toMyVariables(vector<XVariable *> &src, vec<Variable *> &dest) {
            dest.clear();
            for(unsigned int i = 0 ; i < src.size() ; i++)
                dest.push(problem->mapping[src[i]->id]);
            return dest;
        }


        vec<Variable *> &toMyVariables(vector<XVariable *> &src) {
            return toMyVariables(src, vars);
        }


        bool insideGroup;
        int nbMDD;
        vec<Variable *> vars; // Not so beautiful, but efficient...
        vec<int> vals;        // Avoid std::move and save lines...


    public :
        Cosoco::Problem *problem;
        bool optimizationProblem;
        bool invertOptimization; // See Sum objective. If minimize -> Maximize and change sum (only sumGE is supported)
        vec<Variable*> decisionVariables;

        void beginInstance(InstanceType type) override {
            problem = type == CSP ? new Problem("") : new OptimizationProblem("");
            optimizationProblem = type == COP;
            invertOptimization = false;
            nbMDD = 0;
            insideGroup = false;
        }


        void endInstance() override {
            problem->delayedConstruction();
        }


        void buildVariableInteger(string id, int minValue, int maxValue) override {
            problem->createVariable(id, *(new DomainRange(minValue, maxValue)));
        }


        void buildVariableInteger(string id, vector<int> &values) override {
            problem->createVariable(id, *(new DomainValue(vector2vec(values))));
        }


        void beginGroup(string name) override {
            insideGroup = true;
            nbMDD = 0;
        }


        void endGroup() override {
            insideGroup = false;
            nbMDD = 0;

        }


        void beginSlide(string id, bool circular) override {
            insideGroup = true;
            nbMDD = 0;
        }


        void endSlide() override {
            insideGroup = false;
        }

        //--------------------------------------------------------------------------------------
        // Basic constraints
        //--------------------------------------------------------------------------------------

        void buildConstraintExtension(string id, vector<XVariable *> list, vector<vector<int> > &origTuples, bool support, bool hasStar) override {
            if(hasStar && !support)
                throw runtime_error("Extension constraint with star and conflict tuples is not yet supported");
            vec<vec<int> > tuples;
            toMyVariables(list);
            int arity = list.size();
            tuples.growTo(origTuples.size());
            for(unsigned int i = 0 ; i < origTuples.size() ; i++) {
                tuples[i].growTo(arity);
                for(int j = 0 ; j < arity ; j++)
                    tuples[i][j] = origTuples[i][j] == STAR ? STAR : vars[j]->domain.toIdv(origTuples[i][j]);
            }
            FactoryConstraints::createConstraintExtension(problem, id, vars, tuples, support, hasStar);
        }


        void buildConstraintExtension(string id, XVariable *variable, vector<int> &tuples, bool support, bool hasStar) override {
            if(hasStar)
                throw runtime_error("Extension constraint with star is not yet supported");
            FactoryConstraints::createConstraintUnary(problem, id, problem->mapping[variable->id], vector2vec(tuples), support);
        }


        void buildConstraintExtensionAs(string id, vector<XVariable *> list, bool support, bool hasStar) override {
            FactoryConstraints::createConstraintExtensionAs(problem, id, toMyVariables(list), support);
        }


        void buildConstraintIntension(string id, Tree *tree) override {
            vars.clear();
            for(string &s : tree->listOfVariables)
                vars.push(problem->mapping[s]);
            FactoryConstraints::createConstraintIntension(problem, id, tree, vars);
        }


        void buildConstraintPrimitive(string id, OrderType op, XVariable *x, int k, XVariable *y) override {   // x + k op y
            assert(op == LE || op == LT || op == EQ || op == NE);
            vars.clear();
            vars.push(problem->mapping[x->id]);
            vars.push(problem->mapping[y->id]);

            if(k == 0 && (op == EQ || op == NE)) {
                if(op == EQ)
                    FactoryConstraints::createConstraintAllEqual(problem, id, vars);
                else
                    FactoryConstraints::createConstraintAllDiff(problem, id, vars);
                return;
            }
            if(op == LE || op == LT) {
                FactoryConstraints::createConstraintLessThan(problem, id, vars[0], k, vars[1], op == LT);
                return;
            }
            string t = k == 0 ? x->id : "add(" + x->id + "," + std::to_string(k) + ")";
            string tmp;
            if(op == EQ) tmp = "eq(";
            if(op == NE) tmp = "ne(";
            assert(op == EQ || op == NE);
            tmp = tmp + t + "," + y->id + ")";
            buildConstraintIntension(id, new Tree(tmp));
        }


        void buildConstraintPrimitive(string id, OrderType op, XVariable *xx, int k) override {                 // x op k op is <= or >=
            Variable *x = problem->mapping[xx->id];
            vec<int> values;
            for(int v : x->domain)
                if((op == LE && v <= k) || (op == GE && v >= k))
                    values.push(v);
            FactoryConstraints::createConstraintUnary(problem, id, x, values, true);
        }


        void buildConstraintPrimitive(string id, XVariable *xx, bool in, int min, int max) override {          // x in/notin [min,max]
            Variable *x = problem->mapping[xx->id];
            vec<int> values;
            for(int v : x->domain)
                if(min <= v && v <= max)
                    values.push(v);
            FactoryConstraints::createConstraintUnary(problem, id, x, values, in);
        }



        //--------------------------------------------------------------------------------------
        // Language  constraints
        //--------------------------------------------------------------------------------------

        Cosoco::MDD *sameMDDAsPrevious(vec<Variable *> &vars) {
            if(insideGroup && nbMDD++ > 0) { // Check is the MDD is the same
                MDDExtension *ext = dynamic_cast<MDDExtension *>(problem->constraints.last());
                assert(ext != nullptr);
                assert(vars.size() == ext->scope.size());
                int same = true;
                for(int i = 0 ; i < vars.size() ; i++)
                    if(vars[i]->minimum() != ext->scope[i]->minimum()
                       || vars[i]->maximum() != ext->scope[i]->maximum()
                       || vars[i]->domain.maxSize() != ext->scope[i]->domain.maxSize()) {
                        same = false;
                        break;
                    }
                if(same)
                    return ext->mdd;
            }
            return nullptr;
        }


        void buildConstraintMDD(string id, vector<XVariable *> &list, vector<XTransition> &transitions) override {
            toMyVariables(list, vars);
            Cosoco::MDD *mdd = nullptr;
            if((mdd = sameMDDAsPrevious(vars)) != nullptr) {
                FactoryConstraints::createConstraintMDD(problem, id, vars, mdd);
                return;
            }

            vec<XTransition *> trans;
            for(unsigned int i = 0 ; i < transitions.size() ; i++)
                trans.push(&transitions[i]);
            FactoryConstraints::createConstraintMDD(problem, id, vars, trans);
        }


        void buildConstraintRegular(string id, vector<XVariable *> &list, string start, vector<string> &final, vector<XTransition> &transitions) override {
            toMyVariables(list, vars);
            Cosoco::MDD *mdd = nullptr;
            if((mdd = sameMDDAsPrevious(vars)) != nullptr) {
                FactoryConstraints::createConstraintMDD(problem, id, vars, mdd);
                return;
            }

            vec<XTransition *> trans;
            for(unsigned int i = 0 ; i < transitions.size() ; i++)
                trans.push(&transitions[i]);
            FactoryConstraints::createConstraintRegular(problem, id, vars, start, final, trans);


        }
        //--------------------------------------------------------------------------------------
        // Comparison constraints
        //--------------------------------------------------------------------------------------

        void buildConstraintAlldifferent(string id, vector<XVariable *> &list) override {
            FactoryConstraints::createConstraintAllDiff(problem, id, toMyVariables(list));
        }


        virtual void buildConstraintAlldifferentExcept(string id, vector<XVariable *> &list, vector<int> &except) override {
            FactoryConstraints::createConstraintAllDiffExcept(problem, id, toMyVariables(list), except);
        }


        void buildConstraintAlldifferent(string id, vector<Tree *> &trees) override  {
            for(unsigned int i = 0 ; i < trees.size(); i++) {
                for(unsigned int j = i + 1 ; j < trees.size() ; j++) {
                    NodeNE *n = new NodeNE();
                    n->addParameter(trees[i]->root);
                    n->addParameter(trees[j]->root);
                    Tree *tmp = new Tree(n);
                    tmp->prefixe();
                    // Problem with scope
                    for(string s : trees[i]->listOfVariables)
                        tmp->listOfVariables.push_back(s);
                    for(string s : trees[j]->listOfVariables)
                        if(std::find(tmp->listOfVariables.begin(),tmp->listOfVariables.end(), s) == tmp->listOfVariables.end())
                            tmp->listOfVariables.push_back(s);
                    buildConstraintIntension(id, tmp);

                }
            }
        }


        void buildConstraintAlldifferentList(string id, vector<vector<XVariable *> > &origlists) override {
            vec<vec<Variable *> > lists;
            for(unsigned int i = 0 ; i < origlists.size() ; i++) {
                lists.push();
                toMyVariables(origlists[i], lists.last());
            }
            FactoryConstraints::createConstraintAllDiffList(problem, id, lists);
        }


        void buildConstraintAlldifferentMatrix(string id, vector<vector<XVariable *> > &matrix) override {
            // lines
            for(unsigned int i = 0 ; i < matrix.size() ; i++)
                buildConstraintAlldifferent(id, matrix[i]);

            // columns
            for(unsigned int i = 0 ; i < matrix[0].size() ; i++) {
                vector<XVariable *> alldiff;
                for(unsigned int j = 0 ; j < matrix.size() ; j++)
                    alldiff.push_back(matrix[j][i]);
                buildConstraintAlldifferent(id, alldiff);
            }
        }


        void buildConstraintAllEqual(string id, vector<XVariable *> &list) override {
            FactoryConstraints::createConstraintAllEqual(problem, id, toMyVariables(list));
        }


        void buildConstraintNotAllEqual(string id, vector<XVariable *> &list) override {
            FactoryConstraints::createConstraintNotAllEqual(problem, id, toMyVariables(list));
        }


        void buildConstraintOrdered(string id, vector<XVariable *> &list, OrderType order) override {
            vector<int> lengths;
            FactoryConstraints::createConstraintOrdered(problem, id, toMyVariables(list), lengths, order);
        }

        void buildConstraintOrdered(string id, vector<XVariable *> &list, vector<int> &lengths, OrderType order) {
            FactoryConstraints::createConstraintOrdered(problem, id, toMyVariables(list), lengths, order);

        }


        void buildConstraintLex(string id, vector<vector<XVariable *> > &lists, OrderType order) override {
            vec<Variable *> list1, list2;
            for(unsigned int i = 0 ; i < lists.size() - 1 ; i++) {
                toMyVariables(lists[i], list1);
                toMyVariables(lists[i + 1], list2);
                FactoryConstraints::createConstraintLex(problem, id, list1, list2, order);
            }
        }


        void buildConstraintLexMatrix(string id, vector<vector<XVariable *> > &matrix, OrderType order) override {
            // lines
            buildConstraintLex(id, matrix, order);

            //columns
            vector<vector<XVariable *>> tmatrix;
            for(unsigned int i = 0 ; i < matrix[0].size() ; i++) {
                vector<XVariable *> tmp;
                for(unsigned int j = 0 ; j < matrix.size() ; j++)
                    tmp.push_back(matrix[j][i]);
                tmatrix.push_back(tmp);
            }
            buildConstraintLex(id, tmatrix, order);
        }


        //--------------------------------------------------------------------------------------
        // Summing and counting constraints
        //--------------------------------------------------------------------------------------

        void buildConstraintSum(string id, vector<XVariable *> &list, XCondition &xc) override {
            vector<int> coeffs;
            coeffs.assign(list.size(), 1);
            buildConstraintSum(id, list, coeffs, xc);
        }


        void buildConstraintSum(string id, vector<XVariable *> &list, vector<int> &origcoeffs, XCondition &xc) override {
            toMyVariables(list, vars);


            vector2vec(origcoeffs);
            if(xc.operandType == VARIABLE) {
                xc.operandType = INTEGER;
                xc.val = 0;
                vars.push(problem->mapping[xc.var]);
                vals.push(-1);
            }
            if(xc.op != IN) {
                FactoryConstraints::createConstraintSum(problem, id, vars, vals, xc.val, xc.op);
                return;
            }

            // Intervals
            FactoryConstraints::createConstraintSum(problem, id, vars, vals, xc.min, GE);
            FactoryConstraints::createConstraintSum(problem, id, vars, vals, xc.max, LE);
        }


        void buildConstraintSum(string id, vector<XVariable *> &list, vector<XVariable *> &coeffs, XCondition &xc) override {
            string tmp = "add(";
            assert(list.size() == coeffs.size());
            for(unsigned int i = 0 ; i < list.size() ; i++) {
                tmp = tmp + "mul(" + list[i]->id + "," + coeffs[i]->id + ")";
                if(i < list.size() - 1) tmp = tmp + ",";
            }
            if(xc.operandType == VARIABLE) {
                xc.operandType = INTEGER;
                xc.val = 0;
                tmp = tmp + ",neg(" + xc.var + ")";
            }
            tmp = tmp + ")";
            if(xc.op != IN) {
                if(xc.op == EQ) tmp = "eq(" + tmp;
                if(xc.op == NE) tmp = "ne(" + tmp;
                if(xc.op == LE) tmp = "le(" + tmp;
                if(xc.op == LT) tmp = "lt(" + tmp;
                if(xc.op == GE) tmp = "ge(" + tmp;
                if(xc.op == GT) tmp = "gt(" + tmp;
                tmp = tmp + "," + std::to_string(xc.val) + ")";
                XCSP3Core::Tree *tree = new Tree(tmp);
                buildConstraintIntension(id, tree);
                return;
            }

            // Intervals
            buildConstraintIntension(id, new XCSP3Core::Tree("ge(" + tmp + "," + std::to_string(xc.min) + ")"));
            buildConstraintIntension(id, new XCSP3Core::Tree("le(" + tmp + "," + std::to_string(xc.max) + ")"));
        }


        void buildConstraintAtMost(string id, vector<XVariable *> &list, int value, int k) override {
            FactoryConstraints::createConstraintAtMost(problem, id, toMyVariables(list), value, k);
        }


        void buildConstraintAtLeast(string id, vector<XVariable *> &list, int value, int k) override {
            FactoryConstraints::createConstraintAtLeast(problem, id, toMyVariables(list), value, k);
        }


        void buildConstraintExactlyK(string id, vector<XVariable *> &list, int value, int k) override {
            FactoryConstraints::createConstraintExactly(problem, id, toMyVariables(list), value, k);
        }


        void buildConstraintExactlyVariable(string id, vector<XVariable *> &list, int value, XVariable *x) override {
            FactoryConstraints::createConstraintExactlyVariable(problem, id, toMyVariables(list), value, problem->mapping[x->id]);
        }


        void buildConstraintCardinality(string id, vector<XVariable *> &list, vector<int> values, vector<int> &intOccurs, bool closed) override {
            vec<Occurs> occurs;
            for(int o : intOccurs) {
                occurs.push();
                occurs.last().value = o;
                occurs.last().type = OCCURS_INTEGER;
            }
            FactoryConstraints::createConstraintCardinality(problem, id, toMyVariables(list), vector2vec(values), occurs);
        }


        void buildConstraintCardinality(string id, vector<XVariable *> &list, vector<int> values, vector<XVariable *> &varOccurs, bool closed) override {
            vec<Occurs> occurs;
            for(XVariable *xv : varOccurs) {
                occurs.push();
                occurs.last().x = problem->mapping[xv->id];
                occurs.last().type = OCCURS_VARIABLE;
            }
            FactoryConstraints::createConstraintCardinality(problem, id, toMyVariables(list), vector2vec(values), occurs);
        }


        void buildConstraintCardinality(string id, vector<XVariable *> &list, vector<int> values, vector<XInterval> &intervalOccurs, bool closed) override {
            vec<Occurs> occurs;
            for(XInterval &xi : intervalOccurs) {
                occurs.push();
                occurs.last().min = xi.min;
                occurs.last().max = xi.max;
                occurs.last().type = OCCURS_INTERVAL;
            }
            FactoryConstraints::createConstraintCardinality(problem, id, toMyVariables(list), vector2vec(values), occurs);
        }

        //--------------------------------------------------------------------------------------
        // Connection constraints
        //--------------------------------------------------------------------------------------

        string createExpression(string minmax, OrderType op, vector<XVariable *> &list, string value) {
            string o;

            if(op == LE) o = "le";
            if(op == LT) o = "lt";
            if(op == GE) o = "ge";
            if(op == GT) o = "gt";
            if(op == EQ) o = "eq";
            if(op == NE) o = "ne";
            if(op == IN)
                throw runtime_error("not yet supported");

            string tmp = o + "(" + value + "," + minmax + "(";
            for(unsigned int i = 0 ; i < list.size() ; i++)
                tmp = tmp + (i != 0 ? "," : "") + list[i]->id;
            tmp = tmp + "))";
            return tmp;

        }


        virtual void buildConstraintMaximum(string id, vector<XVariable *> &list, XCondition &xc) override {
            toMyVariables(list, vars);
            if(xc.operandType == VARIABLE) {
                if(xc.op == EQ)
                    FactoryConstraints::createConstraintMaximumVariableEQ(problem, id, vars, problem->mapping[xc.var]);
                else
                    buildConstraintIntension(id, new Tree(createExpression("max", xc.op, list, xc.var)));
                return;
            }

            if(xc.operandType == INTEGER) {
                if(xc.op == LE || xc.op == LT) {
                    FactoryConstraints::createConstraintMaximumLE(problem, id, vars, xc.op == LE ? xc.val : xc.val - 1);
                    return;
                }
                if(xc.op == GE || xc.op == GT) {
                    FactoryConstraints::createConstraintMaximumGE(problem, id, vars, xc.op == GE ? xc.val : xc.val + 1);
                    return;
                }
                buildConstraintIntension(id, new XCSP3Core::Tree(createExpression("max", xc.op, list, std::to_string(xc.val))));
                return;
            }
            // Interval
            FactoryConstraints::createConstraintMaximumLE(problem, id, vars, xc.min);
            FactoryConstraints::createConstraintMaximumGE(problem, id, vars, xc.max);
        }


        virtual void buildConstraintMinimum(string id, vector<XVariable *> &list, XCondition &xc) override {
            toMyVariables(list, vars);
            if(xc.operandType == VARIABLE) {
                if(xc.op == EQ)
                    FactoryConstraints::createConstraintMinimumVariableEQ(problem, id, vars, problem->mapping[xc.var]);
                else
                    buildConstraintIntension(id, new XCSP3Core::Tree(createExpression("min", xc.op, list, xc.var)));
                return;
            }

            if(xc.operandType == INTEGER) {
                if(xc.op == LE || xc.op == LT) {
                    FactoryConstraints::createConstraintMinimumLE(problem, id, vars, xc.op == LE ? xc.val : xc.val - 1);
                    return;
                }
                if(xc.op == GE || xc.op == GT) {
                    FactoryConstraints::createConstraintMinimumGE(problem, id, vars, xc.op == GE ? xc.val : xc.val + 1);
                    return;
                }
                buildConstraintIntension(id, new XCSP3Core::Tree(createExpression("min", xc.op, list, std::to_string(xc.val))));
                return;
            }
            ///Interval
            FactoryConstraints::createConstraintMinimumLE(problem, id, vars, xc.min);
            FactoryConstraints::createConstraintMinimumGE(problem, id, vars, xc.max);
        }


        void buildConstraintElement(string id, vector<XVariable *> &list, int startIndex, XVariable *index, RankType rank, int value) override {
            Variable *idx = problem->mapping[index->id];
            FactoryConstraints::createConstraintElementConstant(problem, id, toMyVariables(list), idx, startIndex, value);
        }


        void buildConstraintElement(string id, vector<XVariable *> &list, int startIndex, XVariable *index, RankType rank, XVariable *value) override {
            Variable *idx = problem->mapping[index->id];
            Variable *val = problem->mapping[value->id];
            assert(val != nullptr);
            FactoryConstraints::createConstraintElementVariable(problem, id, toMyVariables(list), idx, startIndex, val);
        }


        void buildConstraintElement(string id, vector<int> &list, int startIndex, XVariable *index, RankType rank, XVariable *value) override {
            vec<Variable *> vars;
            vec<vec<int> > tuples;
            Variable *idx = problem->mapping[index->id];
            Variable *val = problem->mapping[value->id];

            vars.push(idx);
            vars.push(val);

            for(int idv1 : idx->domain) {
                int v = idx->domain.toVal(idv1) - startIndex;
                if(v >= 0 && v < list.size() && val->containsValue(list[v])) {
                    tuples.push();
                    tuples.last().push(idx->domain.toIdv(v + startIndex));
                    tuples.last().push(val->domain.toIdv(list[v]));
                }
            }
            FactoryConstraints::createConstraintExtension(problem, id, vars, tuples, true, false);
        }


        void buildConstraintChannel(string id, vector<XVariable *> &list1, int startIndex1, vector<XVariable *> &list2, int startIndex2) override {
            vec<Variable *> X;
            vec<Variable *> Y;
            toMyVariables(list1, X);
            toMyVariables(list2, Y);
            FactoryConstraints::createContraintChannelXY(problem, id, X, Y, startIndex1, startIndex2);
        }


        void buildConstraintChannel(string id, vector<XVariable *> &list, int startIndex, XVariable *value) override {
            //  return forall(range(list.length), i -> intension(iff(list[i], eq(value, i))));
            int i = 0;
            for(XVariable *x : list) {
                string tmp = "iff(" + x->id + "," + "eq(" + value->id + "," + std::to_string(i) + "))";
                i++;
                buildConstraintIntension(id, new XCSP3Core::Tree(tmp));
            }
        }
        //--------------------------------------------------------------------------------------
        // packing and schedulling constraints
        //--------------------------------------------------------------------------------------


        void buildConstraintNoOverlap(string id, vector<XVariable *> &origins, vector<int> &lengths, bool zeroIgnored) override {
            if(!zeroIgnored)
                throw runtime_error("Nooverlap with zeroIgnored not yet supported");

            toMyVariables(origins);
            for(int i = 0 ; i < vars.size() ; i++)
                for(int j = i + 1 ; j < vars.size() ; j++) {
                    FactoryConstraints::createConstraintNoOverlap(problem, id, vars[i], vars[j], lengths[i], lengths[j]);
                }
        }


        //--------------------------------------------------------------------------------------
        // Instantiation constraint
        //--------------------------------------------------------------------------------------

        void buildConstraintInstantiation(string id, vector<XVariable *> &list, vector<int> &values) override {
            for(unsigned int i = 0 ; i < list.size() ; i++) {
                vec<int> value;
                value.push(values[i]);
                FactoryConstraints::createConstraintUnary(problem, id, problem->mapping[list[i]->id], value, true);
            }
        }

        //--------------------------------------------------------------------------------------
        // Objectives
        //--------------------------------------------------------------------------------------

        virtual void buildObjectiveMinimizeExpression(string expr) override {
            OptimizationProblem *po = static_cast<OptimizationProblem *>(problem);
            po->type = OptimisationType::Minimize;
            string tmp = "le(" + expr + ",0)";
            buildConstraintIntension("objective", new XCSP3Core::Tree(tmp));
            ObjectiveConstraint *oc = dynamic_cast<ObjectiveConstraint *>(problem->constraints.last());
            po->addObjectiveUB(oc, true);
        }


        virtual void buildObjectiveMaximizeExpression(string expr) override {
            OptimizationProblem *po = static_cast<OptimizationProblem *>(problem);
            po->type = OptimisationType::Maximize;
            string tmp = "ge(" + expr + ",0)"; //Fake value
            buildConstraintIntension("objective", new XCSP3Core::Tree(tmp));
            ObjectiveConstraint *oc = dynamic_cast<ObjectiveConstraint *>(problem->constraints.last());
            po->addObjectiveLB(oc, true);
        }


        virtual void buildObjectiveMinimizeVariable(XVariable *x) override {
            buildObjectiveMinimizeExpression(x->id);
        }


        virtual void buildObjectiveMaximizeVariable(XVariable *x) override {
            buildObjectiveMaximizeExpression(x->id);
        }


        void buildObjectiveMinimize(ExpressionObjective type, vector<XVariable *> &list) override {
            vector<int> coeffs;
            coeffs.assign(list.size(), 1);
            buildObjectiveMinimize(type, list, coeffs);
        }


        void buildObjectiveMaximize(ExpressionObjective type, vector<XVariable *> &list) override {
            vector<int> coeffs;
            coeffs.assign(list.size(), 1);
            buildObjectiveMaximize(type, list, coeffs);
        }


        void buildObjectiveMinimize(ExpressionObjective type, vector<XVariable *> &list, vector<int> &origcoeffs) override {
            OptimizationProblem *po = static_cast<OptimizationProblem *>(problem);
            po->type = OptimisationType::Minimize;
            toMyVariables(list, vars);
            switch(type) {
                case EXPRESSION_O:
                    break; // useless, want to discard warning
                case SUM_O : {

                    // Change optim type because LE->GE in Sum
                    po->type = OptimisationType::Maximize;
                    invertOptimization = true;

                    FactoryConstraints::createConstraintSum(problem, "objective", vars, vector2vec(origcoeffs), INT_MAX, LE);
                    ObjectiveConstraint *oc = dynamic_cast<ObjectiveConstraint *>(problem->constraints.last());
                    po->addObjectiveLB(oc, true);
                    break;
                }
                case PRODUCT_O :
                    throw runtime_error("Objective product not yet supported");
                    break;
                case MINIMUM_O : {
                    FactoryConstraints::createConstraintMinimumLE(problem, "objective", vars, INT_MAX);
                    ObjectiveConstraint *oc = dynamic_cast<ObjectiveConstraint *>(problem->constraints.last());
                    po->addObjectiveUB(oc, true);
                    break;
                }
                case MAXIMUM_O : {
                    FactoryConstraints::createConstraintMaximumLE(problem, "objective", vars, INT_MAX);
                    ObjectiveConstraint *oc = dynamic_cast<ObjectiveConstraint *>(problem->constraints.last());
                    po->addObjectiveUB(oc, true);
                    break;
                }
                case NVALUES_O : {
                    FactoryConstraints::createConstraintNValuesLE(problem, "objective", vars, vars.size() + 1);
                    ObjectiveConstraint *oc = dynamic_cast<ObjectiveConstraint *>(problem->constraints.last());
                    po->addObjectiveUB(oc, true);
                    break;
                }
                case LEX_O :
                    throw runtime_error("Objective expression not yet supported");
                    break;
            }
        }


        void buildObjectiveMaximize(ExpressionObjective type, vector<XVariable *> &list, vector<int> &origcoeffs) override {
            OptimizationProblem *po = static_cast<OptimizationProblem *>(problem);
            po->type = OptimisationType::Maximize;
            toMyVariables(list, vars);
            switch(type) {
                case EXPRESSION_O:
                    break; // useless, want to discard warning
                case SUM_O : {
                    FactoryConstraints::createConstraintSum(problem, "objective", vars, vector2vec(origcoeffs), INT_MIN, GE);
                    break;
                }
                case PRODUCT_O :
                    throw runtime_error("Objective product not yet supported");
                    break;
                case MINIMUM_O :
                    FactoryConstraints::createConstraintMinimumGE(problem, "objective", vars, INT_MIN);
                    break;
                case MAXIMUM_O :
                    FactoryConstraints::createConstraintMaximumGE(problem, "objective", vars, INT_MIN);
                    break;
                case NVALUES_O :
                    throw runtime_error("Objective expression not yet supported");
                    break;
                case LEX_O :
                    throw runtime_error("Objective expression not yet supported");
                    break;
            }
            ObjectiveConstraint *oc = dynamic_cast<ObjectiveConstraint *>(problem->constraints.last());
            po->addObjectiveLB(oc, true);
        }


        void buildAnnotationDecision(vector<XVariable *> &list) {
            toMyVariables(list, decisionVariables);
        }
    };


}
#endif //COSOCO_COSOCOCALLBACKS_H
