#include <signal.h>
#include <zlib.h>
#include <sys/resource.h>
#include "utils/System.h"
#include "utils/Options.h"
#include "solver/Solver.h"
#include <vector>
#include<sstream>
#include <optimizer/Optimizer.h>
#include "XCSP3CoreParser.h"
#include "CosocoCallbacks.h"

using namespace Cosoco;
using namespace XCSP3Core;

static bool optimize = false;


std::vector<std::string> &split1(const std::string &s, char delim, std::vector<std::string> &elems) {
    std::stringstream ss(s);
    std::string item;
    while(std::getline(ss, item, delim)) {
        elems.push_back(item);
    }
    return elems;
}


std::vector<std::string> split1(const std::string &s, char delim) {
    std::vector<std::string> elems;
    split1(s, delim, elems);
    return elems;
}


void printStats(AbstractSolver &solver) {
    double cpu_time = cpuTime();
    double mem_used = memUsedPeak();

    solver.printFinalStats();

    if(mem_used != 0) printf("c Memory used           : %.2f MB\n", mem_used);
    printf("c CPU time              : %g s\n", cpu_time);
}


static AbstractSolver *solver;
static Solver *cspSolver;


// Note that '_exit()' rather than 'exit()' has to be used. The reason is that 'exit()' calls
// destructors and may cause deadlocks if a malloc/free function happens to be running (these
// functions are guarded by locks for multithreaded use).
static void SIGINT_exit(int signum) {
    printf("\n");
    printf("c *** INTERRUPTED ***\n");

    if(solver->verbose.verbosity >= 0) {
        printStats(*cspSolver);
        printf("\n");
        if(solver->nbSolutions >= 1)
            printf("c nb Solutions = %d\n", solver->nbSolutions);
        if(optimize) {
            if(solver->hasSolution()) {
                printf("s SATISFIABLE\n");
                printf("c Best value found: %ld\n\n", ((Optimizer *) solver)->bestCost());
                if(solver->displayModels)
                    solver->displayCurrentSolution();
            } else
                printf("c No Bound found!\n\ns UNKNOWN\n");

        } else {
            if(solver->nbSolutions > 0 )
                printf("s SATISFIABLE\n");
            else
                printf("s UNKNOWN");
        }
    }
    std::cout << std::flush;
    exit(1);
}


// Terminate by notifying the solver and back out gracefully. This is mainly to have a test-case
// for this feature of the Solver as it may take longer than an immediate call to '_exit()'.
static void SIGINT_interrupt(int signum) { SIGINT_exit(0); }


//=================================================================================================
// Main:

int main(int argc, char **argv) {


    try {
        CosocoCallbacks cb;

        printf("c\nc This is cosoco 1.12 --  \nc\n");

        setUsageHelp("c USAGE: %s [options] <input-file>\n\n  where input may be either in plain or gzipped XCSP3.\n");



        // Extra options:
        //
        IntOption verb("MAIN", "verb", "Verbosity level (0=silent, 1=some, 2=more, 3=full, 4=fullfull).", 1, IntRange(0, 4));
        IntOption cpu_lim("MAIN", "cpu-lim", "Limit on CPU time allowed in seconds.\n", INT32_MAX,
                          IntRange(0, INT32_MAX));
        IntOption mem_lim("MAIN", "mem-lim", "Limit on memory usage in megabytes.\n", INT32_MAX,
                          IntRange(0, INT32_MAX));
        IntOption nbSolutions("MAIN", "nbsols", "Number of solutions to find", 1, IntRange(0, INT32_MAX));
        BoolOption model("MAIN", "model", "Display models", 0);

        BoolOption lastConflict("SEARCH", "lc", "Last Conflict reasoning", 0);
        BoolOption sticking("SEARCH", "stick", "Sticking Value on heuristic val", 0);
        //BoolOption optimize("SEARCH", "cop", "Run optimizer (needs an objective)", 0);
        BoolOption restarts("SEARCH", "restarts", "Enable restarts", 0);
        BoolOption sat("SEARCH", "sat", "Enable SAT engine", 0);

        BoolOption  pg("OPT", "pg", "Enable progress saving (only after a new solution)", false);
        BoolOption  lns("OPT", "lns", "Enable Large Neighboorhood Seearch (LNS)", false);
        StringOption removeClasses("PARSE", "removeclasses", "Remove special classes when parsing (symmetryBreaking,redundant...)");

        printf("c command line: ");
        for(int i = 0; i < argc; i++)
            printf("%s ", argv[i]);
        printf("\n");
        parseOptions(argc, argv, true);
        Problem *solvingProblem = nullptr;

        double initial_time = cpuTime();

        if(removeClasses != 0) {
            std::vector<std::string> classes = split1(std::string(removeClasses), ',');
            for(std::string c : classes)
                cb.addClassToDiscard(c);
        }

        //XCSP3PrintCallbacks cb;
        try {
            XCSP3CoreParser parser(&cb);
            parser.parse(argv[1]); // parse the input file
            solvingProblem = cb.problem;
            optimize = cb.optimizationProblem;
        }
        catch(exception &e) {
            cout.flush();
            cout  << "c " << e.what()  << endl;
            if(strstr(e.what(), "not yet supported"))
                cout << "s UNSUPPORTED" << endl;
            exit(1);
        }


        // Set limit on CPU-time:
        if(cpu_lim != INT32_MAX) {
            rlimit rl;
            getrlimit(RLIMIT_CPU, &rl);
            if(rl.rlim_max == RLIM_INFINITY || (rlim_t) cpu_lim < rl.rlim_max) {
                rl.rlim_cur = cpu_lim;
                if(setrlimit(RLIMIT_CPU, &rl) == -1)
                    printf("c WARNING! Could not set resource limit: CPU-time.\n");
            }
        }

        // Set limit on virtual memory:
        if(mem_lim != INT32_MAX) {
            rlim_t new_mem_lim = (rlim_t) mem_lim * 1024 * 1024;
            rlimit rl;
            getrlimit(RLIMIT_AS, &rl);
            if(rl.rlim_max == RLIM_INFINITY || new_mem_lim < rl.rlim_max) {
                rl.rlim_cur = new_mem_lim;
                if(setrlimit(RLIMIT_AS, &rl) == -1)
                    printf("c WARNING! Could not set resource limit: Virtual memory.\n");
            }
        }

        Solver *S = new Solver(*solvingProblem);
        if(lastConflict)
            S->addLastConflictReasoning();
        if(sticking)
            S->addStickingValue();
        if(restarts)
            S->addRestart();
        if(sat)
            S->enableSATEngine();

        S->nbWishedSolutions = nbSolutions;

        S->setDecisionVariables(cb.decisionVariables);
        cspSolver = S;

        if(verb >= 2)
            solvingProblem->display();


        if(optimize) {
            printf("c enable optimization\n");
            Optimizer *optimizer = new Optimizer;
            optimizer->setSolver(S);
            if(pg)
                optimizer->addProgressSaving();
            if(lns)
                optimizer->addLNS();
            solver = optimizer;
            optimizer->invertBestCost = cb.invertOptimization;
        } else
            solver = S;


        solver->displayModels = model;
        solver->setVerbosity(verb);

        if(solver->verbose.verbosity > 0) {
            printf("c ========================================[ Problem Statistics ]===========================================\n");
            printf("c |                                                                                                       \n");
        }


        if(solver->verbose.verbosity > 0) {
//           printf("c |  Number of variables:  %12d                                                                   |\n", S.nVars());
//           printf("c |  Number of clauses:    %12d                                                                   |\n", S.nClauses()); }
        }
        double parsed_time = cpuTime();
        if(solver->verbose.verbosity > 0) {
            printf("c |  Parse time        : %12.2f s \n",
                   parsed_time - initial_time);
            printf("c |\n");

            printf("c |               Variables: %d\n",
                   solvingProblem->nbVariables());
            printf("c |            Domain Sizes: %d..%d\n", solvingProblem->minimumDomainSize(), solvingProblem->maximumDomainSize());
            printf("c |\n");
            printf("c |             Constraints: %d\n", solvingProblem->nbConstraints());
            printf("c |                   Arity: %d..%d", solvingProblem->minimumArity(), solvingProblem->maximumArity());
            int nb;
            if((nb = solvingProblem->nbConstraintsOfSize(1)) > 0)
                printf("  -- Unary: %d", nb);
            if((nb = solvingProblem->nbConstraintsOfSize(2)) > 0)
                printf("  -- Binary: %d", nb);
            if((nb = solvingProblem->nbConstraintsOfSize(3)) > 0)
                printf("  -- Ternary: %d", nb);

            printf("\nc | \n");
            printf("c |                   Types: \nc |                          ");

            std::map<std::string, int> typeOfConstraints;
            solvingProblem->nbTypeOfConstraints(typeOfConstraints);
            for(auto &iter : typeOfConstraints) {
                if(iter.first == "Extension")
                    printf("Extension: %d  (nb tuples: %d..%d) -- (shared: %d)\nc |                          ", iter.second,
                           solvingProblem->minimumTuplesInExtension(),
                           solvingProblem->maximumTuplesInExtension(), Extension::nbShared);
                else
                    printf("%s: %d\nc |                          ", iter.first.c_str(), iter.second);

            }

            if(optimize) {
                printf("\n");
                printf("c |               Objective: ");
                Optimizer *o = (Optimizer*)solver;
                ObjectiveConstraint *objective = (o->optimtype == Minimize) ? o->objectiveUB : o->objectiveLB;
                Constraint *c = dynamic_cast<Constraint *>(objective);
                printf("%s %s\n", o->optimtype == Minimize || (o->optimtype == Maximize && o->invertBestCost) ? "Minimize" : "Maximize", c->type.c_str());
            }

        }
        printf("\n");
        printf("c =========================================================================================================\n");

        // Change to signal-handlers that will only notify the solver and allow it to terminate
        // voluntarily:
        signal(SIGINT, SIGINT_interrupt);
        signal(SIGXCPU, SIGINT_interrupt);
        signal(SIGTERM, SIGINT_interrupt);
        signal(SIGABRT, SIGINT_interrupt);

        solver->solve();

        if(solver->verbose.verbosity > 0) {
            if(solver->displayModels == false && solver->nbSolutions >= 1)
                printf("c nb Solutions = %d\n\n", solver->nbSolutions);

            printStats(*solver);
            printf("\n");
        }
        if(solver->nbSolutions == 0)
            printf("s UNSATISFIABLE\n");
        else {
            if(optimize)
                printf("s OPTIMUM FOUND\n");
                else
            printf("s SATISFIABLE\n");
        }

        std::cout << std::flush;

    } catch(OutOfMemoryException &) {
        printf("c =========================================================================================================\n");
        printf("s UNKNOWN\n");
        exit(0);
    }
}
