#include <XCSP3Constants.h>
#include "Optimizer.h"
#include "core/OptimizationProblem.h"

using namespace Cosoco;


Optimizer::Optimizer()
        : AbstractSolver(), boundFound(false), invertBestCost(false), callToSolver(0), solver(nullptr), objectiveLB(nullptr), objectiveUB(nullptr), useDicothomicMode(false), progressSaving(false) {

}


void Optimizer::setSolver(Solver *s) {
    solver = s;
    solver->nbWishedSolutions = 1;
    solver->setVerbosity(0);
    optimtype = (static_cast<OptimizationProblem &>(solver->problem)).type;
    objectiveLB = (static_cast<OptimizationProblem &>(solver->problem)).objectiveLB;
    objectiveUB = (static_cast<OptimizationProblem &>(solver->problem)).objectiveUB;
    assert(objectiveLB != nullptr || objectiveUB != nullptr);
    lower = (objectiveLB != nullptr) ? objectiveLB->minLowerBound() : objectiveUB->minLowerBound();
    upper = (objectiveLB != nullptr) ? objectiveLB->maxUpperBound() : objectiveUB->maxUpperBound();
}


void Optimizer::solve() {
    if(solver == nullptr)
        throw std::runtime_error("Solver is not attached to the optimizer!");

    if(useDicothomicMode)
        solveInDichotomicMode();
    else
        solveInOneDirection();
}


void Optimizer::solveInOneDirection() {
    // Bounds are correctly initialized
    bool firstCall = true;
    vec<int> tuple;
    ObjectiveConstraint *objective = (optimtype == Minimize) ? objectiveUB : objectiveLB;
    Constraint *c = dynamic_cast<Constraint *>(objective);
    assert(objective != nullptr);

    if(invertBestCost)
        verbose.log(NORMAL, "c best=?? lower=%d upper=%d \n", -upper, -lower);
    else
        verbose.log(NORMAL, "c best=?? lower=%d upper=%d \n", lower, upper);

    solver->checkSolution = true;
    while(status == RUNNING) {
        if(optimtype == Minimize)
            objective->updateBound(upper);
        else
            objective->updateBound(lower);

        solver->solve();
        callToSolver++;
        solver->fullBacktrack();
        solver->reinitializeConstraints();
        if(solver->hasASolution()) {
            nbSolutions++;
            c->extractConstraintTupleFromInterpretation(solver->lastSolution, tuple);
            best = objective->computeScore(tuple);

            // Store solution in order to avoid a signal
            bestSolution.begin(bestCost());
            for(int i = 0; i < solver->lastSolution.size(); i++)
                bestSolution.appendTo(solver->problem.variables[i]->_name, solver->problem.variables[i]->useless ? STAR : solver->lastSolution[i]);
            bestSolution.end();

            if(progressSaving) {
                vec<int> idvalues;
                idvalues.growTo(solver->problem.nbVariables());
                for(Variable *x : solver->problem.variables)
                    idvalues[x->idx] = solver->problem.variables[x->idx]->domain[0];
                ((ForceIdvs *) solver->heuristicVal)->setIdValues(idvalues);
            }

            if(lns) {
                HeuristicVarLNS *hlns = (HeuristicVarLNS *) solver->heuristicVar;
                vec<int> idvalues;
                idvalues.growTo(solver->problem.nbVariables());
                for(Variable *x : solver->problem.variables)
                    idvalues[x->idx] = solver->problem.variables[x->idx]->domain[0];
                hlns->setIdvs(idvalues);
                fragmentSizeInLNS = solver->problem.nbVariables() / 5;
                hlns->maxConflicts = 1000;
                verbose.log(NORMAL, "c New Solution: create fragment with %d vars\n", fragmentSizeInLNS);
                hlns->createFragment(fragmentSizeInLNS);
            }
            if(optimtype == Minimize) upper = best - 1; else lower = best + 1;
            std::cout << "o " << bestCost() << std::endl;
            if(callToSolver % 2 == 1) {
                if(invertBestCost)
                    verbose.log(NORMAL, "c %4d calls to solver: best=%d lower=%d upper=%d \n", callToSolver, bestCost(), -upper, -lower);
                else
                    verbose.log(NORMAL, "c %4d calls to solver: best=%d lower=%d upper=%d \n", callToSolver, bestCost(), lower, upper);
            }
        } else {
            if(lns && fragmentSizeInLNS > 0) {
                HeuristicVarLNS *hlns = (HeuristicVarLNS *) solver->heuristicVar;
                if(hlns->fragmentIsUnsat()) {
                    verbose.log(NORMAL, "c Fragment is unsat\n");
                    fragmentSizeInLNS = fragmentSizeInLNS == 0 ? 0 : fragmentSizeInLNS - 1;
                    hlns->maxConflicts = 1000;
                    hlns->addFragmentToNogoods();
                } else {
                    hlns->maxConflicts *= 1.2;
                    verbose.log(NORMAL, "c Add more conflicts %d\n", hlns->maxConflicts);
                }

                verbose.log(NORMAL, "c Create fragment with %d vars\n", fragmentSizeInLNS);
                hlns->createFragment(fragmentSizeInLNS);
                continue;
            }
            status = OPTIMUM;
            if(firstCall)
                std::cout << "c Optimization: first call of Solver is UNSAT !\n" << "s UNSATISFIABLE\n";
        }
        firstCall = false;
    }
    if(nbSolutions > 0)
        verbose.log(NORMAL, "c\nc Optimum found         : %d\nc\n", bestCost());
    if(displayModels)
        displayCurrentSolution();

}


void Optimizer::solveInDichotomicMode() {
    // Bounds are correctly initialized
    assert(false);
    /*bool firstCall = true;
    long middle;
    //best = (optimtype == Minimize) ? upper : lower;
    Verbose::log(NORMAL, "c best=?? lower=%d middle=%d upper=%d \n", lower, middle, upper);
    while (status == RUNNING) {
        middle = (lower + upper) / 2;
        assert(lower <= upper);
        objective->updateBound(middle);
        solver->solve();
        solver->fullBacktrack(true);
        if (solver->hasASolution()) {
            best = objective->computeScore(solver->lastSolution);
            Verbose::log(NORMAL, "c best=?? lower=%d middle=%d upper=%d \n", best, lower, middle, upper);
            if (optimtype == Minimize) upper = best - 1; else lower = best + 1;
        } else {
            if (firstCall)
                throw std::string("Optimization: first call of Solver is UNSAT !\n");
            // We are sure limit is initialized !
            // we do not care about middle because first call is sat !

            if (optimtype == Minimize) lower = middle + 1; else upper = middle - 1;
        }
        if (lower > upper) { // This is the end
            status = OPTIMUM;
        }
        firstCall = false;

    }
    Verbose::log(NORMAL, "c Optimum found : %d\n", best);
     */

}


void Optimizer::displayCurrentSolution() {
    if(bestSolution.exists())
        bestSolution.display();
}


void Optimizer::printFinalStats() {
    printf("c call to solver        : %d\nc\n", callToSolver);
    solver->printFinalStats();
}
