#ifndef SOLVER_H
#define    SOLVER_H

#include <solver/restarts/Restart.h>
#include <satwrapper/SatWrapper.h>
#include <solver/observers/ObserverDomainReduction.h>
#include "mtl/Vec.h"
#include "AbstractSolver.h"
#include "heuristics/variables/HeuristicVar.h"
#include "heuristics/values/HeuristicVal.h"
#include "core/Problem.h"
#include "mtl/SparseSetOfVariables.h"
#include "observers/ObserverConflict.h"
#include "observers/ObserverDecision.h"
#include "restarts/Restart.h"
#include "DecisionMarker.h"
#include "satwrapper/SatWrapper.h"

namespace Cosoco {
#define NBSTATS 7
    enum GlobalStats {
        rootPropagations,
        uselessFilterCalls,
        restarts,
        nogoods,
        SATPropagations,
        SATConflicts,
        SATLearnts
    };

#define NBLOCALSTATS 4
    enum OneRunStats {
        maxDepth,
        minDepth,
        sumDepth,
        nbConflicts
    };

    class Restart;
    class DecisionMarker;
    class SatWrapper;

    class Solver : public AbstractSolver {
    public :
        Problem &problem;                                   // The problem to solve
        int nbWishedSolutions;                              // The number of solutions to find
        vec<int> lastSolution;                              // The last solution found
#ifdef COMPARESOLUTIONS
        vec<vec<int> > allSolutions;
#endif
        // -- Main Stats ------------------------------------------------------------------------
        uint64_t decisions, conflicts, propagations;
        uint64_t filterCalls;


        // -- Minor Stats -----------------------------------------------------------------------
        vec<uint64_t> statistics;                           // global statistics
        vec<int> localstatistics;                           // local statistics: cleared at each restart
        int displayStatsEveryConflicts;                     // Display GlobalStats every conflicts


        // -- Minor fields ----------------------------------------------------------------------
        double seed;                                        // A seed value
        bool checkSolution;                                 // Check solution or not
        bool filterCallIsUsefull;                           // Check if a call to filter call is usefull or not


        // -- Search ----------------------------------------------------------------------------
        vec<Variable *> trail;                               // the trail of variables
        // Each level stores the SET variables of variables touched
        vec<int> trail_lim;                                  // the limit for each level
        SparseSetOfVariables unassignedVariables;            // The set of unassigned variables;
        SparseSetOfVariables decisionVariables;
        DecisionMarker *decisionMarker;                      // The current branch of the search tree

        // -- Heuristics ------------------------------------------------------------------------
        HeuristicVar *heuristicVar;                          // The heuristic to choose variables
        HeuristicVal *heuristicVal;                          // The heuristic to choose values
        Restart *restart;                                    // The restart strategy

        // -- Propagations ----------------------------------------------------------------------
        SparseSetOfVariables queue;                          // Propagation queue
        Constraint *currentFilteredConstraint;               // The constraint that is filtered

        // -- Observers ----------------------------------------------------------------------
        vec<ObserverConflict *> observersConflict;           // Classes listen for conflict
        vec<ObserverDecision *> observersDecision;           // Classes listen for decisions
        vec<ObserverDomainReduction* >observersDomainReduction; // Classes listen for domain reduction
        unsigned long timestamp;                             // Current timestamp

        // Optionnal techniques
        bool nogoodsFromRestarts;                            // Generate nogoods from restarts
        SatWrapper *satWrapper;


        // --------------------------------------------------------------------------------------
        // Construction and initialisation
        // --------------------------------------------------------------------------------------

        Solver(Problem &p);
        void addLastConflictReasoning();
        void addStickingValue();
        void addRestart();
        void enableSATEngine();


        // --------------------------------------------------------------------------------------
        // Main methods
        // --------------------------------------------------------------------------------------

        void solve();                                        // The main function
        void search();                                       // The search function
        bool manageSolution();                               // Return true if the search is finished

        // --------------------------------------------------------------------------------------
        // Decision methods
        // --------------------------------------------------------------------------------------
        void newDecision(Variable *x, int idv);              // Assign a variable
        Variable *decisionVariableAtLevel(int lvl);          // Return the decision variable
        int decisionLevel();                                 // Return the decision level
        bool isAssigned(Variable *x);                        // Return true if x is assigned


        // --------------------------------------------------------------------------------------
        // Backtrack methods
        // --------------------------------------------------------------------------------------

        void doRestart();
        void backtrack();                                    // Backtrack to previous level
        void fullBacktrack(bool all = false);                // Backtrack to the root, if all=true remove everything (including root AC)
        void handleFailure(Variable *x, int idv);            // manage backtrack until fix point
        void handleFailure(int level);                       // Manage backtrack to  this level (at least)
        void reinitializeConstraints();                      // Force constraints to be propagated

        // --------------------------------------------------------------------------------------
        // Propagate methods
        // --------------------------------------------------------------------------------------

        void addToQueue(Variable *x);                             // Add a variable to queue (side effect if used directly))
        Variable *pickInQueue();                                  // Pick a var in the prop queue
        Constraint *propagate(bool startWithSATEngine = false);   // Propagate the Queue
        Constraint *propagateComplete();                          // fill the queue and propagate everything
        bool isGACGuaranted();                                    // Return trus if GAC is ensured


        // --------------------------------------------------------------------------------------
        // Domain variable modifications (add/remove values)
        // --------------------------------------------------------------------------------------

        bool delVal(Variable *x, int v);                     // Use it to delete variable values
        bool delIdv(Variable *x, int idv);                   // Use it to delete variable values

        bool assignToVal(Variable *x, int v);                // Use it to assign variable value
        bool assignToIdv(Variable *x, int idv);              // Use it to assign variable value


        // --------------------------------------------------------------------------------------
        // Multiple values removal
        // --------------------------------------------------------------------------------------

        bool delValuesGreaterOrEqualThan(Variable *x, int v);
        bool delValuesLowerOrEqualThan(Variable *x, int v);
        bool delValuesInDomain(Variable *x, Domain &d);
        bool delValuesNotInDomain(Variable *x, Domain &d);
        bool changeDomain(Variable *x, SparseSet &newIdvalues);


        // --------------------------------------------------------------------------------------
        // Observers methods
        // --------------------------------------------------------------------------------------

        void addObserverConflict(ObserverConflict *oc);      // oc wants to know when a conflict is reached
        void notifyConflict(Constraint *c);                  // Notify classes that a conflict is reached

        void addObserverDecision(ObserverDecision *od);      // od wants to know when a decision is added or deleted
        void notifyNewDecision(Variable *x);                 // Notify classes that x is a new decision
        void notifyDeleteDecision(Variable *x,int v);        // Notify classes that x=v is not a decision,
        void notifyFullBactrack();                           // Notify classes that we remove all facts

        void addObserverDomainReduction(ObserverDomainReduction *odr);
        void notifyDomainReduction(Variable *x, int idv);    //
        void notifyDomainAssignment(Variable *x, int idv);    //

        // --------------------------------------------------------------------------------------
        // minor methods
        // --------------------------------------------------------------------------------------

        virtual void displayCurrentSolution();
        virtual void printFinalStats();                 // The final stats to print

        void displayHeaderCurrentSearchSpace();
        void displayCurrentSearchSpace();

        void updateStatisticsWithNewConflict();
        bool hasASolution() { return lastSolution.size() > 0; }
        void displayTrail();

        void interrupt();          // Trigger a (potentially asynchronous) interruption of the solver.

        void setDecisionVariables(vec<Variable*> &vars);      // Set the decision variables

    };


};

#endif	/* SOLVER_H */

